# -*- coding: utf-8 -*-
#
# Author: Ingelrest François (Athropos@gmail.com)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA

import gtk, gui, os.path, tools

from tools   import consts
from gettext import gettext as _


mCurrentDir = consts.dirUsr


def openFiles(parent, title, filterPatterns={}):
    """
        Return a list of files, None if any
        The format of filter must be {'Name1': ['filter1', 'filter2'], 'Name2': ['filter3'] ... }
    """
    global mCurrentDir

    dlg = gtk.FileChooserDialog(title, parent, gtk.FILE_CHOOSER_ACTION_OPEN, (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL, gtk.STOCK_OPEN, gtk.RESPONSE_OK))

    # Set some specific options and restore the previously used directory
    dlg.set_select_multiple(True)
    dlg.set_current_folder(mCurrentDir)

    # Add filters
    for name,patterns in filterPatterns.iteritems():
        filter = gtk.FileFilter()
        filter.set_name(name)
        map(filter.add_pattern, patterns)
        dlg.add_filter(filter)

    files = None
    if dlg.run() == gtk.RESPONSE_OK:
        files = dlg.get_filenames()

    # Save the directory for the next time and destroy the dialog
    mCurrentDir = dlg.get_current_folder()
    dlg.destroy()

    return files


def openDirectory(parent, title):
    """ Return a directory, None if any """
    global mCurrentDir

    dlg = gtk.FileChooserDialog(title, parent, gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER, (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL, gtk.STOCK_OPEN, gtk.RESPONSE_OK))

    # Set some specific options and restore the previously used directory
    dlg.set_select_multiple(False)
    dlg.set_current_folder(mCurrentDir)

    directory = None
    if dlg.run() == gtk.RESPONSE_OK:
        directory = dlg.get_filename()

    # Save the directory for the next time and destroy the dialog
    mCurrentDir = dlg.get_current_folder()
    dlg.destroy()

    return directory


def save(parent, title, defaultFile, defaultDir=None):
    """ Return a filename chosen by the user, None if any """
    global mCurrentDir

    dlg = gtk.FileChooserDialog(title, parent, gtk.FILE_CHOOSER_ACTION_SAVE, buttons=(gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL, gtk.STOCK_SAVE, gtk.RESPONSE_OK))

    # Set some specific options and restore the previously used directory
    dlg.set_current_name(defaultFile)
    dlg.set_do_overwrite_confirmation(True)

    if defaultDir is None: dlg.set_current_folder(mCurrentDir)
    else:                  dlg.set_current_folder(defaultDir)

    file = None
    if dlg.run() == gtk.RESPONSE_OK:
        file = dlg.get_filename()

    # Save the directory for the next time and destroy the dialog
    mCurrentDir = dlg.get_current_folder()
    dlg.destroy()

    return file


# --== Selecting a path ==--


spDlg            = None
spName           = None
spPath           = None
spBtnOk          = None
spForbiddenNames = []
spForbiddenChars = []


def __spOnOpen(btn):
    """ Let the user choose a folder, and fill the corresponding field in the dialog """
    path = openDirectory(spDlg, _('Choose a folder'))
    if path is not None:
        spPath.set_text(path)


def __spOnTextChanged(entry):
    """ Enable/disable the OK button based on the content of the text fields """
    spBtnOk.set_sensitive(spName.get_text() != '' and spPath.get_text() != '')


def __spCheckResponse(dlg, response, *args):
    """ If response is gtk.RESPONSE_OK, check that the values are correct """
    if response == gtk.RESPONSE_OK:
        name = spName.get_text()
        path = spPath.get_text()

        if not os.path.isdir(path):
            gui.errorMsgBox(spDlg, _('This path does not exist'), _('Please select an existing directory.'))
            spDlg.stop_emission('response')
        elif name in spForbiddenNames:
            gui.errorMsgBox(spDlg, _('The name is incorrect'), _('This name is not allowed.\nPlease use another one.'))
            spDlg.stop_emission('response')
        else:
            for c in name:
                if c in spForbiddenChars:
                    gui.errorMsgBox(spDlg, _('The name is incorrect'), _('The character %s is not allowed.\nPlease use another name.') % c)
                    spDlg.stop_emission('response')
                    break


def selectPath(parent, title, forbiddenNames=[], forbiddenChars=[]):
    """ Return a tuple (name, path) or None if the user cancelled the dialog box """
    global spDlg, spName, spPath, spBtnOk, spForbiddenNames, spForbiddenChars

    if spDlg is None:
        spTree  = tools.loadGladeFile('SelectPath.glade')
        spDlg   = spTree.get_widget('dlg')
        spName  = spTree.get_widget('txt-name')
        spPath  = spTree.get_widget('txt-path')
        spBtnOk = spTree.get_widget('btn-ok')
        # Connect handlers
        spTree.get_widget('btn-open').connect('clicked', __spOnOpen)
        spName.connect('changed', __spOnTextChanged)
        spPath.connect('changed', __spOnTextChanged)
        spDlg.connect('response', __spCheckResponse)

    spName.set_text('')
    spPath.set_text('')
    spName.grab_focus()
    spDlg.set_title(title)
    spBtnOk.set_sensitive(False)
    spDlg.set_transient_for(parent)
    spForbiddenNames = forbiddenNames
    spForbiddenChars = forbiddenChars

    if spDlg.run() == gtk.RESPONSE_OK:
        spDlg.hide()
        return (spName.get_text(), spPath.get_text())

    spDlg.hide()
    return None
