# -*- coding: utf-8 -*-
#
# Author: Ingelrest François (Athropos@gmail.com)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA

import cgi, gobject, gtk.glade, modules, os.path

from gui        import help, infoMsgBox, errorMsgBox, extListview
from tools      import consts, prefs
from gettext    import gettext as _
from gui.window import Window


(
    ROW_ENABLED,    # True if the module is currently enabled
    ROW_TEXT,       # Name and description of the module
    ROW_ICON,       # An icon indicating whether the module is configurable
    ROW_NAME,       # Name of the module
    ROW_UNLOADABLE  # True if the module can be disabled
) = range(5)


class Preferences:
    """ Allow the user to load/unload/configure the modules """


    def __init__(self):
        """ Constructor """
        self.window = None


    def show(self):
        """ Show the dialog box """
        if self.window is None:
            self.window = Window('Preferences.glade', 'vbox1', __name__, _('Preferences'), 390, 350)
            # List of modules
            toggleRdr = gtk.CellRendererToggle()
            columns   = (('',   [(toggleRdr, gobject.TYPE_BOOLEAN)],             ROW_ENABLED,    False),
                         ('',   [(gtk.CellRendererText(), gobject.TYPE_STRING)], ROW_TEXT,       True),
                         ('',   [(gtk.CellRendererPixbuf(), gtk.gdk.Pixbuf)],    ROW_ICON,       False),
                         (None, [(None, gobject.TYPE_STRING)],                   ROW_NAME,       False),
                         (None, [(None, gobject.TYPE_BOOLEAN)],                  ROW_UNLOADABLE, False))

            self.list = extListview.ExtListView(columns, sortable=False, useMarkup=True)
            self.list.set_headers_visible(False)
            self.list.addColumnAttribute(0, toggleRdr, 'activatable', ROW_UNLOADABLE)
            toggleRdr.connect('toggled', self.onModuleToggled)
            self.window.getWidget('scrolledwindow1').add(self.list)
            self.fillList()
            # GTK handlers
            self.window.getWidget('btn-prefs').connect('clicked', self.onPreferences)
            self.window.getWidget('btn-help').connect('clicked', self.onHelp)
            self.window.getWidget('btn-close').connect('clicked', lambda btn: self.window.hide())
            self.list.get_selection().connect('changed', self.onSelectionChanged)

        if not self.window.isVisible():
            self.window.getWidget('btn-prefs').set_sensitive(False)
            self.window.getWidget('btn-close').grab_focus()
            self.list.unselectAll()

        self.window.show()


    def fillList(self):
        """ Fill the list of modules """
        rows = []
        for (name, data) in sorted([m for m in modules.getModules() if m[1][modules.MOD_CONFIGURABLE] or not m[1][modules.MOD_MANDATORY]]):
            if data[modules.MOD_CONFIGURABLE] and data[modules.MOD_INSTANCE]: icon = consts.icoBtnPrefs
            else:                                                             icon = None

            text = '<b>%s</b>\n<small>%s</small>' % (cgi.escape(name), cgi.escape(data[modules.MOD_DESC]))
            rows.append([data[modules.MOD_INSTANCE] is not None, text, icon, name, not data[modules.MOD_MANDATORY]])

        self.list.replaceContent(rows)


    # --== GTK handlers ==--


    def onModuleToggled(self, renderer, path):
        """ A module has been enabled/disabled """
        row = self.list.getRow(path)

        if row[ROW_ENABLED]:
            modules.unload(row[ROW_NAME])
        else:
            try:                             modules.load(row[ROW_NAME])
            except modules.LoadException, e: errorMsgBox(self.window, _('Unable to load this module.'), str(e))

        self.fillList()


    def onHelp(self, btn):
        """ Show a small help message box """
        helpDlg = help.HelpDlg(_('Modules'))
        helpDlg.addSection(_('Description'), _('This dialog box shows the list of available modules, which are small pieces of code that add some functionnalities to the application. You can enable/disable a module by checking/unchecking the check box in front of it. Note that some modules (e.g., the File Explorer) cannot be disabled.'))
        helpDlg.addSection(_('Configuring a Module'), _('When a module may be configured, a specific icon is displayed on the right of the corresponding line. To configure a module, simply select it and then click on the "Preferences" button on the bottom of the dialog box. Note that configuring a module is only possible when it is enabled.'))
        helpDlg.show(self.window)


    def onSelectionChanged(self, selection):
        """ Decide whether the new selection may be configured """
        if self.list.getSelectedRowsCount() != 1:
            self.window.getWidget('btn-prefs').set_sensitive(False)
        else:
            row = self.list.getFirstSelectedRow()
            self.window.getWidget('btn-prefs').set_sensitive(row[ROW_ENABLED] and row[ROW_ICON] is not None)


    def onPreferences(self, btn):
        """ Configure the selected module """
        modules.configure(self.list.getFirstSelectedRow()[ROW_NAME], self.window)


# --== Global functions ==--


__instance = Preferences()

def show():
    """ Show the preferences dialog box """
    __instance.show()
