# -*- coding: utf-8 -*-
#
# Author: Ingelrest François (Athropos@gmail.com)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA

import flac, mp3, ogg, os, playlist, tools

from tools.consts import UNKNOWN_TRACKNUMBER, UNKNOWN_TITLE, UNKNOWN_ARTIST, UNKNOWN_ALBUM, UNKNOWN_LENGTH


# The format of the list returned by the readInfo() and getTracks() functions
(
    NFO_NUM,  # Track number
    NFO_TIT,  # Title
    NFO_ART,  # Artist
    NFO_ALB,  # Album
    NFO_LEN,  # Length in seconds
    NFO_FIL   # Full path to the file
) = range(6)


# Supported formats with associated modules
mFormats = {'.mp3': mp3, '.ogg': ogg, '.flac': flac}


def isSupported(file):
    """
        Return True if the given file is a supported format
    """
    return os.path.splitext(file)[1].lower() in mFormats


def getSupportedFormats():
    """
        Return a list of all formats from which tags can be extracted
        This list does not account for the installed gstreamer packages
    """
    return ['*' + ext for ext in mFormats.iterkeys()]


def readInfo(files):
    """
        Given a list of files, return a list such as:
        [[FILE1, TRACKNUMBER1, TITLE1, ARTIST1, ALBUM1, LENGTH1], [FILE2, TRACKNUMBER2, TITLE2, ARTIST2, ALBUM2, LENGTH2], ...]
    """
    info = []
    for file in files:
        try:
            module = mFormats[os.path.splitext(file)[1].lower()]
            info.append(module.readInfo(file))
        except KeyError:
            tools.log.logger.warning('Unknown file format: %s' % file)
            info.append([UNKNOWN_TRACKNUMBER, UNKNOWN_TITLE, UNKNOWN_ARTIST, UNKNOWN_ALBUM, UNKNOWN_LENGTH, file])

    return info


def __cmpTracks(track):
    """
        Used to sort tracks
    """
    return track[NFO_NUM]


def getTracks(files):
    """
        Return all the tracks contained in the list of files:
            * For directory, load all media files (playable files and playlists)
            * For a playlist, load the tracks it contains
        The content of playlists is never sorted
    """
    tracks = []

    # Directories
    for directory in [file for file in files if os.path.isdir(file)]:
        mediaFiles, playlists = [], []
        for (filename, fullPath) in tools.listDir(directory):
            if os.path.isfile(fullPath):
                if isSupported(filename):            mediaFiles.append(fullPath)
                elif playlist.isSupported(filename): playlists.append(fullPath)

        tracks.extend(sorted(readInfo(mediaFiles), key=__cmpTracks))

        for pl in playlists:
            tracks.extend(readInfo(playlist.load(pl)))

    # Files
    mediaFiles = sorted(readInfo([file for file in files if os.path.isfile(file) and isSupported(file)]), key=__cmpTracks)
    tracks.extend(mediaFiles)

    # Playlists
    for pl in [file for file in files if os.path.isfile(file) and playlist.isSupported(file)]:
        tracks.extend(readInfo(playlist.load(pl)))

    return tracks
