# -*- coding: utf-8 -*-
#
# Author: Ingelrest François (Athropos@gmail.com)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA

import gtk, gtk.glade, gui.window, media, modules, os, tools, urllib

from cgi     import escape
from gui     import fileChooser, help, questionMsgBox, extTreeview, extListview
from tools   import consts, prefs
from media   import playlist
from gettext import gettext as _
from os.path import isdir, isfile
from gobject import idle_add, TYPE_STRING, TYPE_INT


# Module configuration
MOD_NAME            = _('File Explorer')
MOD_DESC            = _('Navigate through your file system')
MOD_IS_MANDATORY    = True
MOD_IS_CONFIGURABLE = True


# Default preferences
PREFS_DEFAULT_MEDIA_FOLDERS     = {_('Home'): consts.dirUsr}    # List of media folders that are used as roots for the file explorer
PREFS_DEFAULT_SHOW_HIDDEN_FILES = False                         # True if hidden files should be shown


# The format of a row in the treeview
(
    ROW_PIXBUF,    # Item icon
    ROW_NAME,      # Item name
    ROW_TYPE,      # The type of the item (e.g., directory, file)
    ROW_FULLPATH   # The full path to the item
) = range(4)


# The possible types for a node of the tree
(
    TYPE_DIR,   # A directory
    TYPE_PLIST, # A playlist
    TYPE_FILE,  # A media file
    TYPE_NONE   # A fake item, used to display a '+' in front of a directory when needed
) = range(4)


class FileExplorer(modules.Module):
    """ This explorer lets the user browse the disk from a given root directory (e.g., ~/, /) """


    def __init__(self, wTree):
        """ Constructor """
        modules.Module.__init__(self)
        modules.register(self, (modules.MSG_EVT_APP_STARTED, modules.MSG_EVT_APP_QUIT, modules.MSG_EVT_EXPLORER_CHANGED))
        # Attributes
        self.currRoot        = None
        self.popupMenu       = None
        self.cfgWindow       = None
        self.mediaFolders    = prefs.get(__name__, 'media-folders',     PREFS_DEFAULT_MEDIA_FOLDERS)
        self.showHiddenFiles = prefs.get(__name__, 'show-hidden-files', PREFS_DEFAULT_SHOW_HIDDEN_FILES)
        # Create the tree
        columns = (('',   [(gtk.CellRendererPixbuf(), gtk.gdk.Pixbuf), (gtk.CellRendererText(), TYPE_STRING)], True),
                   (None, [(None, TYPE_INT)],                                                                  False),
                   (None, [(None, TYPE_STRING)],                                                               False))

        self.tree = extTreeview.ExtTreeView(columns)
        self.tree.setIsDraggableFunc(self.isDraggable)
        self.tree.setDNDSources([consts.DND_TARGETS[consts.DND_DAP_URI]])
        # Scroll window
        self.scrolled = gtk.ScrolledWindow()
        self.scrolled.add(self.tree)
        self.scrolled.set_shadow_type(gtk.SHADOW_IN)
        self.scrolled.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        self.scrolled.show()
        # GTK handlers
        self.tree.connect('drag-data-get',              self.onDragDataGet)
        self.tree.connect('exttreeview-row-expanded',   self.onRowExpanded)
        self.tree.connect('exttreeview-button-pressed', self.onButtonPressed)


    # Helper method
    def __cmpRows(self, row): return row[ROW_NAME].lower()


    def setShowHiddenFiles(self, show):
        """ Show/hide hidden files """
        if show != self.showHiddenFiles:
            self.showHiddenFiles = not self.showHiddenFiles
            prefs.set(__name__, 'show-hidden-files', self.showHiddenFiles)
            self.refresh(self.tree)


    def playSelection(self, replace):
        """ Replace/extend the tracklist """
        path = [row[ROW_FULLPATH] for row in self.tree.getSelectedRows()]

        if replace: modules.postMsg(modules.MSG_CMD_TRACKLIST_SET, {'files': path, 'playNow': True})
        else:       modules.postMsg(modules.MSG_CMD_TRACKLIST_ADD, {'files': path})


    def showPopupMenu(self, button, time, path):
        """ Show a popup menu """
        if self.popupMenu is None:
            wTree                = tools.loadGladeFile('FileExplorerMenu.glade')
            self.popupMenu       = wTree.get_widget('menu-popup')
            self.popupItemAdd    = wTree.get_widget('item-add')
            self.popupItemPlay   = wTree.get_widget('item-play')
            self.popupItemHidden = wTree.get_widget('item-hidden')
            # Connect handlers
            self.popupItemAdd.connect('activate', lambda widget: self.playSelection(False))
            self.popupItemPlay.connect('activate', lambda widget: self.playSelection(True))
            wTree.get_widget('item-refresh').connect('activate', lambda widget: self.refresh(self.tree))
            wTree.get_widget('item-hidden').connect('toggled', lambda item: self.setShowHiddenFiles(item.get_active()))

        self.popupItemHidden.set_active(self.showHiddenFiles)
        self.popupItemAdd.set_sensitive(path is not None and self.tree.getItem(path, ROW_PIXBUF) != consts.icoDir)
        self.popupItemPlay.set_sensitive(path is not None and self.tree.getItem(path, ROW_PIXBUF) != consts.icoDir)
        self.popupMenu.popup(None, None, None, button, time)


    def getDirContent(self, directory):
        """ Return a tuple of sorted rows (directories, playlists, mediaFiles) for the given directory """
        playlists, mediaFiles, directories = [], [], []

        if isdir(directory):
            for (file, fullPath) in tools.listDir(directory, self.showHiddenFiles):
                if isdir(fullPath):
                    directories.append((consts.icoDir, unicode(file, errors='replace'), TYPE_DIR, fullPath))
                elif isfile(fullPath):
                    if media.isSupported(file):      mediaFiles.append((consts.icoMediaFile, unicode(file, errors='replace'), TYPE_FILE, fullPath))
                    elif playlist.isSupported(file): playlists.append((consts.icoMediaFile, unicode(file, errors='replace'), TYPE_PLIST, fullPath))

            directories.sort(key=self.__cmpRows)
            playlists.sort(key=self.__cmpRows)
            mediaFiles.sort(key=self.__cmpRows)

        return (directories, playlists, mediaFiles)


    def updateChildren(self, tree, parent):
        """ This generator updates (e.g., icon, fake child for directories) all children of the given parent """
        for child in tree.iterChildren(parent):
            if tree.getItem(child, ROW_TYPE) != TYPE_DIR:
                yield False

            directory = tree.getItem(child, ROW_FULLPATH)
            if not os.access(directory, os.R_OK | os.X_OK):
                continue

            hasContent, hasMediaContent = False, False
            for (file, fullPath) in tools.listDir(directory, self.showHiddenFiles):
                if isdir(fullPath):
                    hasContent = True
                elif isfile(fullPath) and (media.isSupported(file) or playlist.isSupported(file)):
                    hasContent, hasMediaContent = True, True
                    break

            # Append/remove children if needed
            if hasContent and tree.getNbChildren(child) == 0:      tree.appendRow((consts.icoDir, '', TYPE_NONE, ''), child)
            elif not hasContent and tree.getNbChildren(child) > 0: tree.removeAllChildren(child)

            # Change the icon based on whether the folder contains some media content
            if hasMediaContent: tree.setItem(child, ROW_PIXBUF, consts.icoMediaDir)
            else:               tree.setItem(child, ROW_PIXBUF, consts.icoDir)

            yield True

        yield False


    def exploreDir(self, directory, parent):
        """ List the content of the given directory and append it to the tree as a child of parent """
        directories, playlists, mediaFiles = self.getDirContent(directory)

        self.tree.appendRows(directories, parent)
        self.tree.appendRows(playlists,   parent)
        self.tree.appendRows(mediaFiles,  parent)

        if len(directories) != 0:
            idle_add(self.updateChildren(self.tree, parent).next)


    def refresh(self, tree, treePath=None):
        """ Refresh the tree, starting from treePath """
        if treePath is None: directory = self.mediaFolders[self.currRoot]
        else:                directory = tree.getItem(treePath, ROW_FULLPATH)

        directories, playlists, mediaFiles = self.getDirContent(directory)

        disk = directories
        disk.extend(playlists)
        disk.extend(mediaFiles)

        # Find whether some files have been added/removed
        diskIndex, childIndex = 0, 0
        while diskIndex < len(disk):
            file    = disk[diskIndex]
            rowPath = tree.getChild(treePath, childIndex)

            if rowPath is None:
                tree.appendRow(file, treePath)
                continue

            cmpResult = cmp(self.__cmpRows(tree.getRow(rowPath)), self.__cmpRows(file))
            if cmpResult < 0:
                tree.removeRow(rowPath)
            else:
                if cmpResult > 0:
                    tree.insertRowBefore(file, treePath, rowPath)
                diskIndex  += 1
                childIndex += 1

        # Some tree rows may be left: it means that the corresponding files no longer exist
        while childIndex < tree.getNbChildren(treePath):
            tree.removeRow(tree.getChild(treePath, childIndex))

        # Refresh also expanded rows
        for child in tree.iterChildren(treePath):
            if tree.row_expanded(child):
                idle_add(self.refresh, tree, child)

        # Update nodes' appearance
        if len(directories) != 0:
            idle_add(self.updateChildren(tree, treePath).next)


    def isDraggable(self):
        """ Determine whether the selected rows can be dragged """
        for row in self.tree.iterSelectedRows():
            if row[ROW_PIXBUF] == consts.icoDir:
                return False
        return True


    # --== GTK handlers ==--


    def onButtonPressed(self, tree, event, path):
        """ A mouse button has been pressed """
        if event.button == 3:
            self.showPopupMenu(event.button, event.time, path)
        elif event.button == 1 and event.type == gtk.gdk._2BUTTON_PRESS and path is not None:
            if   tree.getItem(path, ROW_PIXBUF) != consts.icoDir: self.playSelection(True)
            elif tree.row_expanded(path):                         tree.collapse_row(path)
            else:                                                 tree.expand_row(path, False)


    def onRowExpanded(self, tree, path):
        """ Populate the expanded row if needed """
        child = tree.getChild(path, 0)
        if tree.getItem(child, ROW_TYPE) == TYPE_NONE:
            self.exploreDir(tree.getItem(path, ROW_FULLPATH), path)
            tree.removeRow(child)


    def onDragDataGet(self, tree, context, selection, info, time):
        """ Provide information about the data being dragged """
        selection.set('text/uri-list', 8, ' '.join(['file://' + urllib.pathname2url(row[ROW_FULLPATH]) for row in tree.getSelectedRows()]))


    def onAppStarted(self):
        """ Add all media folders to the explorer Module """
        for (name, path) in self.mediaFolders.iteritems():
            modules.postMsg(modules.MSG_CMD_EXPLORER_ADD, {'modName': MOD_NAME, 'expName': name, 'icon': None, 'widget': self.scrolled})


   # --== Message handler ==--


    def handleMsg(self, msg, params):
        """ Handle messages sent to this module """
        if msg == modules.MSG_EVT_EXPLORER_CHANGED and params['modName'] == MOD_NAME and self.currRoot != params['expName']:
            self.currRoot = params['expName']
            self.tree.clear()
            self.exploreDir(self.mediaFolders[self.currRoot], None)
            if len(self.tree) != 0:
                self.tree.scroll_to_cell(0)
        elif msg == modules.MSG_EVT_APP_STARTED:
            idle_add(self.onAppStarted)
        elif msg == modules.MSG_EVT_APP_QUIT:
            pass


    # --== Configuration ==--


    def configure(self, parent):
        """ Show the configuration dialog """
        if self.cfgWindow is None:
            self.cfgWindow = gui.window.Window('FileExplorer.glade', 'vbox1', __name__, MOD_NAME, 370, 400)
            # Create the list of media folders
            txtRdr  = gtk.CellRendererText()
            pixRdr  = gtk.CellRendererPixbuf()
            columns = ((None, [(txtRdr, TYPE_STRING)],                           0, False),
                       ('',   [(pixRdr, gtk.gdk.Pixbuf), (txtRdr, TYPE_STRING)], 2, False))

            self.cfgList = extListview.ExtListView(columns, sortable=False, useMarkup=True)
            self.cfgList.set_headers_visible(False)
            self.cfgWindow.getWidget('scrolledwindow1').add(self.cfgList)
            self.cfgWindow.getWidget('btn-remove').set_sensitive(False)
            # Connect handlers
            self.cfgList.connect('key-press-event', self.onCfgKeyboard)
            self.cfgList.get_selection().connect('changed', self.onCfgSelectionChanged)
            self.cfgWindow.getWidget('btn-add').connect('clicked', self.onAddMediaFolder)
            self.cfgWindow.getWidget('btn-remove').connect('clicked', lambda btn: self.removeSelectedMediaFolders(self.cfgList))
            self.cfgWindow.getWidget('btn-ok').connect('clicked', self.onBtnOk)
            self.cfgWindow.getWidget('btn-cancel').connect('clicked', lambda btn: self.cfgWindow.hide())
            self.cfgWindow.getWidget('btn-help').connect('clicked', self.onHelp)

        if not self.cfgWindow.isVisible():
            self.fillMFList()
            self.cfgWindow.getWidget('chk-hidden').set_active(self.showHiddenFiles)
            self.cfgWindow.getWidget('btn-ok').grab_focus()
        self.cfgWindow.show()


    def onAddMediaFolder(self, btn):
        """ Let the user add a new media folder to the list """
        result = fileChooser.selectPath(self.cfgWindow, MOD_NAME, self.mediaFolders.keys())

        if result is not None:
            name, path = result
            self.mediaFolders[name] = path
            prefs.set(__name__, 'media-folders', self.mediaFolders)
            self.fillMFList()
            modules.postMsg(modules.MSG_CMD_EXPLORER_ADD, {'modName': MOD_NAME, 'expName': name, 'icon': None, 'widget': self.scrolled})


    def fillMFList(self):
        """ Fill the list of currently used media folders """
        rows = [(name, consts.icoBtnDir, '<b>%s</b>\n<small>%s</small>' % (escape(name), escape(path)))
                for name, path in sorted(self.mediaFolders.iteritems())]
        self.cfgList.replaceContent(rows)


    def removeSelectedMediaFolders(self, list):
        """ Remove the selected media folder """
        if list.getSelectedRowsCount() == 1:
            remark   = _('You will be able to add this root folder again later on if you wish so.')
            question = _('Remove the selected entry?')
        else:
            remark   = _('You will be able to add these root folders again later on if you wish so.')
            question = _('Remove all selected entries?')

        if questionMsgBox(self.cfgWindow, question, '%s %s' % (_('Your media files will not be deleted.'), remark)) == gtk.RESPONSE_YES:
            for row in self.cfgList.getSelectedRows():
                name = row[0]
                modules.postMsg(modules.MSG_CMD_EXPLORER_REMOVE, {'modName': MOD_NAME, 'expName': name})
                del self.mediaFolders[name]
                prefs.set(__name__, 'media-folders', self.mediaFolders)
                if self.currRoot == name:
                    self.currRoot = None
            self.cfgList.removeSelectedRows()


    def onCfgKeyboard(self, list, event):
        """ Remove the selection if possible """
        if gtk.gdk.keyval_name(event.keyval) == 'Delete':
            self.removeSelectedMediaFolders(list)


    def onCfgSelectionChanged(self, selection):
        """ The selection has changed """
        self.cfgWindow.getWidget('btn-remove').set_sensitive(selection.count_selected_rows() != 0)


    def onBtnOk(self, btn):
        """ The user has clicked on the OK button """
        self.cfgWindow.hide()
        self.setShowHiddenFiles(self.cfgWindow.getWidget('chk-hidden').get_active())


    def onHelp(self, btn):
        """ Display a small help message box """
        helpDlg = help.HelpDlg(MOD_NAME)
        helpDlg.addSection(_('Description'),
                           _('This module allows to browse the files on your drives.'))
        helpDlg.addSection(_('Usage'),
                           _('At least one root folder must be added to browse your files. This folder then becomes the root of the '
                             'file explorer tree in the main window.'))
        helpDlg.show(self.cfgWindow)
